<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Config\MigrationActionConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Config\MigrationMappingConfig;
use Dedi\SyliusMigrationPlugin\Service\Logger\MigrationLogger;
use Dedi\SyliusMigrationPlugin\Service\MigrationExecutorInterface;
use Doctrine\DBAL\Exception\DriverException;
use Psr\Log\LogLevel;

class MigrationExecutor implements MigrationExecutorInterface
{
    private MigrationConfig $config;
    private MigrationActionConfig $migrationActionConfig;
    private MigrationMappingConfig $mappingRegistry;
    private ReaderInterface $reader;
    private WriterInterface $writer;
    private MigrationLogger $migrationLogger;

    public function __construct(
        MigrationConfig $config,
        MigrationActionConfig $migrationActionConfig,
        MigrationMappingConfig $mappingRegistry,
        ReaderInterface $reader,
        WriterInterface $writer,
        MigrationLogger $migrationLogger
    ) {
        $this->config = $config;
        $this->migrationActionConfig = $migrationActionConfig;
        $this->mappingRegistry = $mappingRegistry;
        $this->reader = $reader;
        $this->writer = $writer;
        $this->migrationLogger = $migrationLogger;
    }

    public function execute(bool $commit): void
    {
        $commit = $commit && $this->config->isCommit();
        if (!$commit) {
            $this->migrationLogger->log(LogLevel::INFO, 'Running in dry mode');
            $this->defuse();
        }

        $this->migrationActionConfig->executePreMigrationActions($commit);

        $numberOfTables = count($this->mappingRegistry->getTableDefinitions());
        foreach ($this->mappingRegistry->getTableDefinitions() as $index => $tableDefinition) {
            $this->migrationLogger->info(
                $tableDefinition,
                sprintf(
                    'Import from %s to %s [%s/%s]',
                    '' === $tableDefinition->getSourceTableName() ? 'SQL Query' : $tableDefinition->getSourceTableName(),
                    $tableDefinition->getTargetTableName(),
                    $index + 1,
                    $numberOfTables
                )
            );

            if (!$this->writer->tableExists($tableDefinition, $tableDefinition->getTargetTableName())) {
                continue;
            }

            $this->writer->handleExtra($tableDefinition);

            if (!$this->reader->tableExists($tableDefinition, $tableDefinition->getSourceTableName())) {
                continue;
            }

            $arrData = $this->reader->read($tableDefinition);

            $count = 0;
            foreach ($arrData as $data) {
                try {
                    $this->writer->write($tableDefinition, $data);
                } catch (DriverException $exception) {
                    $this->migrationLogger->handleException($exception, $tableDefinition, $data);
                }
                ++$count;

                if ($count === $this->config->getDebugLimitStatements()) {
                    break;
                }
            }
        }

        $this->migrationActionConfig->executePostMigrationActions($commit);
    }

    private function defuse()
    {
        $this->reader->defuse();
        $this->writer->defuse();
    }
}
