<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Definition\ColumnMappingDefinition;
use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Doctrine\DBAL\Exception;
use Doctrine\DBAL\Query\QueryBuilder;
use LogicException;
use Symfony\Component\Serializer\NameConverter\CamelCaseToSnakeCaseNameConverter;

class Writer extends AbstractSqlProcessor implements WriterInterface
{
    public function write(TableMappingDefinition $tableMappingDefinition, $data): void
    {
        $destQueryBuilder = $this->connection->createQueryBuilder();

        $dataExists = $this->handleSearch($tableMappingDefinition, $data);

        if (true === $dataExists) {
            $destQueryBuilder
                ->update($tableMappingDefinition->getTargetTableName());
        } else {
            $destQueryBuilder
                ->insert($tableMappingDefinition->getTargetTableName());
        }

        foreach ($tableMappingDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
            $function = $dataExists ? 'set' : 'setValue';

            call_user_func(
                [$destQueryBuilder, $function],
                $columnMappingDefinition->getTargetColumName(),
                sprintf(':%s', $columnMappingDefinition->getTargetColumName())
            );

            $destQueryBuilder
                ->setParameter($columnMappingDefinition->getTargetColumName(), $this->resolveMappingValue($columnMappingDefinition, $data));
        }

        if (true === $dataExists) {
            $this->addSearchPartsToQuery($tableMappingDefinition, $destQueryBuilder, $data);
        }

        $this->logger->debug($tableMappingDefinition, $destQueryBuilder->getSQL(), $data);

        if ($this->config->isCommit()) {
            $destQueryBuilder->execute();
        }
    }

    public function handleExtra(TableMappingDefinition $tableMappingDefinition): void
    {
        $nameConverter = new CamelCaseToSnakeCaseNameConverter();
        foreach ($tableMappingDefinition->getExtra() as $extra => $params) {
            $functionName = $nameConverter->denormalize($extra);
            call_user_func([$this, $functionName], $tableMappingDefinition, $params['value']);
        }
    }

    /**
     * @throws \Doctrine\DBAL\Driver\Exception
     * @throws Exception
     */
    private function resolveMappingValue(ColumnMappingDefinition $columnMappingDefinition, array $data)
    {
        $value = $data[$columnMappingDefinition->getSourceColumnName()] ?? null;

        if (null !== $columnMappingDefinition->getValue()) {
            return $columnMappingDefinition->getValue();
        }

        if (null !== $columnMappingDefinition->getAlias()) {
            $value = $data[$columnMappingDefinition->getAlias()] ?? null;
        }

        if (null !== $columnMappingDefinition->getSubQuerySql()) {
            $value = $this->connection->executeQuery($columnMappingDefinition->getSubQuerySql(), $data)->fetchOne();

            if (false === $value) {
                $value = null;
            }
        }

        if (null !== $columnMappingDefinition->getTransformer()) {
            $transformer = $this->transformerRegistry->getTransformer($columnMappingDefinition->getTransformer());
            $value = $transformer->execute($value);
        }

        return $value;
    }

    private function legacyColumnName(TableMappingDefinition $tableMappingDefinition, $legacyColumnName)
    {
        if (!$tableMappingDefinition->hasColumnMapping($legacyColumnName)) {
            $this->logger->warning($tableMappingDefinition, sprintf('You should define a field mapping for you legacy column %s', $legacyColumnName));
        }

        $tableDescription = $this
            ->connection
            ->executeQuery(sprintf('DESCRIBE `%s`', $tableMappingDefinition->getTargetTableName()))->fetchAllAssociative();

        foreach ($tableDescription as $description) {
            if ($description['Field'] == $legacyColumnName) {
                return;
            }
        }

        $sql = sprintf('ALTER TABLE `%s` ADD COLUMN `%s` VARCHAR(255) NULL', $tableMappingDefinition->getTargetTableName(), $legacyColumnName);
        $this->logger->debug($tableMappingDefinition, $sql);

        if ($this->config->isCommit()) {
            $this
                ->connection
                ->executeQuery($sql);
        }
    }

    /**
     * @throws Exception
     */
    private function handleSearch(TableMappingDefinition $tableMappingDefinition, $data): bool
    {
        $result = null;
        if ($tableMappingDefinition->hasSearchMapping()) {
            $searchQueryBuilder = $this->connection->createQueryBuilder();
            $searchQueryBuilder
                ->select('*')
                ->from($tableMappingDefinition->getTargetTableName());

            $this->addSearchPartsToQuery($tableMappingDefinition, $searchQueryBuilder, $data);

            $this->logger->debug($tableMappingDefinition, $searchQueryBuilder->getSQL(), $data);
            $result = $searchQueryBuilder->execute()->fetchAllAssociative();
        }

        return null !== $result && count($result) > 0;
    }

    private function addSearchPartsToQuery(TableMappingDefinition $tableMappingDefinition, QueryBuilder $queryBuilder, $data): void
    {
        foreach ($tableMappingDefinition->getSearchMapping() as $searchMappingDefinition) {
            if (!array_key_exists($searchMappingDefinition->getValueColum(), $data)) {
                throw new LogicException(sprintf('Column %s does not exists in data', $searchMappingDefinition->getValueColum()));
            }

            $pattern = $this->resolveSearchType($searchMappingDefinition->getType());
            $value = $this->prepareDataForSearch($searchMappingDefinition->getType(), $data[$searchMappingDefinition->getValueColum()]);

            $queryBuilder
                ->andWhere(sprintf($pattern, $searchMappingDefinition->getSearchColumn(), $searchMappingDefinition->getValueColum()))
                ->setParameter($searchMappingDefinition->getValueColum(), $value);
        }
    }

    private function resolveSearchType(string $type)
    {
        switch ($type) {
            case 'comparison':
                return '%s = :%s';
            case 'not_comparison':
                return '%s != :%s';
            case 'like':
                return '%s LIKE :%s';
            case 'is_null':
                return '%s IS NULL';
            default:
                throw new LogicException('Unsupported type %s', $type);
        }
    }

    private function prepareDataForSearch(string $type, $value)
    {
        switch ($type) {
            case 'comparison':
                return $value;
            case 'like':
                return sprintf('%%%s%%', $value);
            default:
                throw new LogicException('Unsupported type %s', $type);
        }
    }
}
